//---------------------------------------------------------------------------
// GPGee - GNU Privacy Guard Explorer Extension
// GPGeeVerifyDecrypt.cpp - Verification form code
// Copyright 2005, Kurt Fitzner <kfitzner@excelcia.org>
//---------------------------------------------------------------------------
// This file is part of GPGee.
//
// GPGee is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// GPGee is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
//---------------------------------------------------------------------------
#include <vcl.h>
#pragma hdrstop

#include "TConfiguration.h"
#include "GPGeeExceptions.h"
#include "GPGeePassPhrase.h"
#include "GPGeeVerifyDecrypt.h"
//---------------------------------------------------------------------------
#pragma package(smart_init)
#pragma resource "*.dfm"
TformGPGeeVerifyDecrypt *formGPGeeVerifyDecrypt;
extern TConfiguration *GPGeeConfig;
//---------------------------------------------------------------------------


//---------------------------------------------------------------------------
// Form constructor
//
__fastcall TformGPGeeVerifyDecrypt::TformGPGeeVerifyDecrypt(TComponent* Owner): TForm(Owner)
{
  gpgme_error_t err;
  AnsiString sConfigString;

  ctxGPG = NULL;
  cacheSecret = NULL;
  cachePublic = NULL;

  Cancel = false;

  // Initialize the GPGME library main and keycache contexts
  err = gpgme_new(&ctxGPG);
  if (err)
    throw EGPGMEError("GPGee encountered an error while initialising the GPGME library: " + String(gpgme_strerror(err)));
/*
  err = gpgme_keycache_new (&cachePublic, ctxGPG);
  if (err)
    throw EGPGMEError("GPGee encountered an error while initialising the public key cache: " + String(gpgme_strerror(err)));
  err = gpgme_keycache_init(cachePublic, NULL, false);
  if (err)
    throw EGPGMEError("GPGee encountered an error while caching your GnuPG public keys: " + String(gpgme_strerror(err)));
  gpgme_set_cache_ctx(ctxGPG, cachePublic); */
  err = gpgme_keycache_new (&cacheSecret, ctxGPG);
  if (err)
    throw EGPGMEError("GPGee encountered an error while initialising the secret key cache: " + String(gpgme_strerror(err)));
  err = gpgme_keycache_init(cacheSecret, NULL, true);
  if (err)
    throw EGPGMEError("GPGee encountered an error while caching your GnuPG secret keys: " + String(gpgme_strerror(err)));
  gpgme_set_passphrase_cb(ctxGPG, PassphraseCallback, this);

  // Set GPGME's settings to match GPGee's configuration
  gpgme_control(ctxGPG, GPGME_CTRL_FORCETRUST, GPGeeConfig->Values["Force Trust"]?1:0);
  sConfigString = GPGeeConfig->Values["GPG Program"];
  if (!sConfigString.IsEmpty())
    gpgme_control(ctxGPG, GPGME_CTRL_GPGPROGRAM, (int)sConfigString.c_str());
  sConfigString = GPGeeConfig->Values["Options File"];
  if (!sConfigString.IsEmpty())
    gpgme_control(ctxGPG, GPGME_CTRL_OPTFILE, (int)sConfigString.c_str());
  sConfigString = GPGeeConfig->Values["Keyring"];
  if (!sConfigString.IsEmpty())
    gpgme_control(ctxGPG, GPGME_CTRL_KEYRING, (int)sConfigString.c_str());
  sConfigString = GPGeeConfig->Values["Secret Keyring"];
  if (!sConfigString.IsEmpty())
    gpgme_control(ctxGPG, GPGME_CTRL_SECRETKEYRING, (int)sConfigString.c_str());
}  // __fastcall TformGPGeeVerifyDecrypt::TformGPGeeVerifyDecrypt(TComponent* Owner): TForm(Owner)
//---------------------------------------------------------------------------


//---------------------------------------------------------------------------
// Death and destruction
//
__fastcall TformGPGeeVerifyDecrypt::~TformGPGeeVerifyDecrypt()
{
  for (int n=0; n<lvFileList->Items->Count; n++) {
    delete lvFileList->Items->Item[n]->Data;
    lvFileList->Items->Item[n]->Data = NULL;
  }  // for (int n=0; n<lvFileList->Items->Count; n++)
  if (cacheSecret)
    gpgme_keycache_release(cacheSecret);
  if (cachePublic)
    gpgme_keycache_release(cachePublic);
  if (ctxGPG)
    gpgme_release(ctxGPG);
}  // __fastcall TformGPGeeVerifyDecrypt::~TformGPGeeVerifyDecrypt()
//---------------------------------------------------------------------------


//---------------------------------------------------------------------------
// OnShow event handler for the form
//
void __fastcall TformGPGeeVerifyDecrypt::FormShow(TObject *Sender)
{
  TListItem *FileItem;

  FileNames->Sort();
  lvFileList->Clear();
  for (int n=0; n<FileNames->Count; n++) {
    VerifyDecryptInfo_t *ItemInfo = new VerifyDecryptInfo_t;

    FileItem = lvFileList->Items->Add();
    FileItem->Caption = MinimizeName(FileNames->Strings[n],lvFileList->Font,lvFileList->Column[0]->Width-20);
    FileItem->Data = ItemInfo;

    ItemInfo->FileName = FileNames->Strings[n];
    ItemInfo->FileType = IdentifyFile(ItemInfo->FileName);
    if (ItemInfo->FileType == GPG_UNKNOWN) {
      ItemInfo->ResultString = "GPGee could not identify what operation to perform on this file.  This is either because it is a GnuPG file that is not supported by GPGee, or it isn't a GnuPG file at all.";
      FileItem->StateIndex = GPGEE_STATE_INVALID;
      FileItem->SubItems->Add("Unknown");
    } else {
      FileItem->StateIndex = GPGEE_STATE_PENDING;
      FileItem->SubItems->Add((ItemInfo->FileType==GPG_MESSAGE)?"Message":"Sig");
    }  // if (ItemInfo->FileType == GPG_UNKNOWN)
  }  // for (int n=0; n<FileNames->Count; n++)
  lvFileList->SetFocus();
  StartTimer->Enabled = true;
}  // void __fastcall TformGPGeeVerifyDecrypt::FormShow(TObject *Sender)
//---------------------------------------------------------------------------


//---------------------------------------------------------------------------
// Set the default modes and show the form
//
int __fastcall TformGPGeeVerifyDecrypt::ShowModalWithDefaults(TStringList *FileNames)
{
  this->FileNames = FileNames;
  return ShowModal();
}  // int __fastcall TformGPGeeSignEncrypt::ShowModalWithDefaults(bool bSign, bool bEncrypt)
//---------------------------------------------------------------------------


//---------------------------------------------------------------------------
// StartTimer OnTimer event - this is used to allow the form time to get
// painted before we start into the operations.
//
void __fastcall TformGPGeeVerifyDecrypt::StartTimerTimer(TObject *Sender)
{
  gpgme_error_t err;
  gpgme_sig_t sig = NULL;

  StartTimer->Enabled = false;
  Application->ProcessMessages();
  for (int n=0; n<lvFileList->Items->Count; n++) {
    TListItem *Item = lvFileList->Items->Item[n];
    VerifyDecryptInfo_t *ItemInfo = (VerifyDecryptInfo_t *)Item->Data;
    AnsiString DataFileName = ItemInfo->FileName.SubString(0,ItemInfo->FileName.Length()-4);
    Item->Selected = true;
    if (Item->StateIndex == GPGEE_STATE_INVALID)
      continue;
    Item->StateIndex = GPGEE_STATE_INPROGRESS;
    Application->ProcessMessages();
    if (Cancel) {
      Item->StateIndex = GPGEE_STATE_INVALID;
      ItemInfo->ResultString = "Operation canceled";
      continue;
    }  // if (Cancel)
    switch (ItemInfo->FileType) {
      case GPG_MESSAGE: {
        gpgme_op_flags_t OpFlags;
        char KeyID[17];

        ExtraPassphraseInfo = ItemInfo->FileName;
        if (FileExists(DataFileName)) {
          AnsiString FileOverwrite = GPGeeConfig->Values["File Overwrite"];
          if (FileOverwrite.AnsiCompareIC("never") == 0 || FileOverwrite.AnsiCompareIC("ask") == 0 &&
              Application->MessageBox(("Overwrite file \"" + DataFileName + "\"?").c_str(), "Confirm Overwrite", MB_YESNO) == IDNO) {
            Item->StateIndex = GPGEE_STATE_INVALID;
            Application->ProcessMessages();
            ItemInfo->ResultString = "Output file exists and overwriting was forbidden.";
            continue;
          }  // if (can't overwrite)
        }  // if (FileExists(DataFileName))
        err = gpgme_op_file_decrypt(ctxGPG, ("\"" + ItemInfo->FileName + "\"").c_str(), ("\"" + DataFileName + "\"").c_str());
        gpgme_decrypt_get_status(ctxGPG, KeyID, &OpFlags);
        if (err == GPGME_No_Error) {
          ItemInfo->ResultString = "Successfully decrypted and written to file \"" + DataFileName + "\"";
          gpgme_decrypt_get_sig_ctx(ctxGPG, &sig);
          if (sig)
            Item->StateIndex = InterpretSignatureResult(sig, ItemInfo->ResultString);
          else
            Item->StateIndex = GPGEE_STATE_SUCCESS;
        } else {
          ItemInfo->ResultString = String(gpgme_strerror(err));
          Item->StateIndex = GPGEE_STATE_FAILURE;
        }  // if (err == GPGME_No_Error)
        Application->ProcessMessages();
        if (OpFlags & GPGME_OPFLAG_NOSECKEY)
          ItemInfo->ResultString += " Keyring does not have the secret key (0x" + String(KeyID) + ") needed to decrypt this message.";
        if (OpFlags & GPGME_OPFLAG_BADARMOR)
          ItemInfo->ResultString += " Message has bad ASCII armor.";
        if (OpFlags & GPGME_OPFLAG_BADMDC)
          ItemInfo->ResultString += " WARNING: This message's modification detection code does not match the message.  This means it is likely that the message has been changed or corrupted since the time it was encrypted.";
        break;
      }  // case GPG_MESSAGE:
      case GPG_SIGNATURE: {
        if (!FileExists(DataFileName)) {
          Item->StateIndex = GPGEE_STATE_INVALID;
          Application->ProcessMessages();
          ItemInfo->ResultString = "GPGee was unable to find the file (" + DataFileName + ") that this signature is for.";
          break;
        }  // if (!FileExists(DataFileName))
        err = gpgme_op_file_verify(ctxGPG, GPGME_SIG_MODE_DETACH, &sig, ("\"" + ItemInfo->FileName + "\"").c_str(), ("\"" + DataFileName + "\"").c_str());
        Item->StateIndex = InterpretSignatureResult(sig, ItemInfo->ResultString, err);
        break;
      }  // case GPG_SIGNATURE:
    }  // switch (ItemInfo->FileType)
    if (sig) {
      gpgme_sig_release(sig);
      sig = NULL;
    }  // if (sig)
  }  // for (int n=0; n<lvFileList->Items->Count; n++)
  lvFileList->Selected = lvFileList->Items->Item[0];
  btnOkCancel->Caption = "Ok";
}  // void __fastcall TformGPGeeVerifyDecrypt::StartTimerTimer(TObject *Sender)
//---------------------------------------------------------------------------


//---------------------------------------------------------------------------
// Look at the result of checking a signature and return the appropriate
// state code.  ResultString is also appended with a text description of the
// results.
//
int __fastcall TformGPGeeVerifyDecrypt::InterpretSignatureResult(gpgme_sig_t sig, AnsiString &ResultString, gpgme_error_t err)
{
  bool FurtherInfo = true;;
  AnsiString SigKeyId;

  int retval;

  if (!ResultString.IsEmpty())
    ResultString += "\r\n";

  SigKeyId = String(gpgme_sig_get_string_attr(sig, GPGME_ATTR_KEYID));
  switch (gpgme_sig_get_ulong_attr(sig, 0, GPGME_ATTR_VALIDITY)) {
    case GPGME_SIG_STAT_NONE:
      ResultString += "No signature - " + String(gpgme_strerror(err));
      retval = GPGEE_STATE_FAILURE;
      FurtherInfo = false;
      break;
    case GPGME_SIG_STAT_GOOD:
      ResultString += "Good signature ";
      retval = GPGEE_STATE_SUCCESS;
      break;
    case GPGME_SIG_STAT_BAD:
      ResultString += "BAD signature ";
      retval = GPGEE_STATE_FAILURE;
      break;
    case GPGME_SIG_STAT_NOKEY:
      ResultString += "Could not verify as the keyring does not have the correct key (0x" + SigKeyId + ")";
      retval = GPGEE_STATE_FAILURE;
      FurtherInfo = false;
      break;
    case GPGME_SIG_STAT_NOSIG:
      ResultString += "No signature was detected - could not verify.";
      retval = GPGEE_STATE_FAILURE;
      FurtherInfo = false;
      break;
    case GPGME_SIG_STAT_ERROR:
      ResultString += "Verify error - " + String(gpgme_strerror(err));
      retval = GPGEE_STATE_FAILURE;
      FurtherInfo = false;
      break;
    case GPGME_SIG_STAT_E_GOOD:
      ResultString += "Good expired-key signature ";
      retval = GPGEE_STATE_PARTIAL;
      break;
    case GPGME_SIG_STAT_R_GOOD:
      ResultString += "Good REVOKED-KEY signature ";
      retval = GPGEE_STATE_PARTIAL;
      break;
  }  // switch (gpgme_sig_get_ulong_attr(sig, 0, GPGME_ATTR_VALIDITY))
  if (FurtherInfo)
    ResultString += "made by" + String(gpgme_sig_get_string_attr(sig, GPGME_ATTR_USERID)) + " (" + KeyTypeToString((gpgme_pk_cipher_t)gpgme_sig_get_ulong_attr(sig, 0, GPGME_ATTR_ALGO)) +"/0x" + String(gpgme_sig_get_string_attr(sig, GPGME_ATTR_KEYID)).SubString(9,8) + ") on " + DateTimeToStr(GPGToDateTime(gpgme_sig_get_ulong_attr(sig, 0, GPGME_ATTR_CREATED))) + ".";
  return retval;
}
//---------------------------------------------------------------------------


//---------------------------------------------------------------------------
// OnChange event handler for the listview component.  Display the extended
// information about a file in the memo field when the user selects it.
//
void __fastcall TformGPGeeVerifyDecrypt::lvFileListChange(TObject *Sender, TListItem *Item, TItemChange Change)
{
  VerifyDecryptInfo_t *Info = (VerifyDecryptInfo_t *)Item->Data;
  if (Change == ctState && Info)
    memoResults->Lines->Text = Info->ResultString;
}  // void __fastcall TformGPGeeVerifyDecrypt::lvFileListChange(TObject *Sender, TListItem *Item, TItemChange Change)
//---------------------------------------------------------------------------


//---------------------------------------------------------------------------
// User pressed the ok/cancel button - If we are currently performing
// operations then we set the cancel flag.  If operations are completed,
// we simply close the dialog.
//
void __fastcall TformGPGeeVerifyDecrypt::btnOkCancelClick(TObject *Sender)
{
  if (btnOkCancel->Caption == "Ok")
    ModalResult = mrOk;
  else
    Cancel = true;
}  // void __fastcall TformGPGeeVerifyDecrypt::btnOkCancelClick(TObject *Sender)
//---------------------------------------------------------------------------


//---------------------------------------------------------------------------
// User pressed the help button - pull up the correct help page
void __fastcall TformGPGeeVerifyDecrypt::btnHelpClick(TObject *Sender)
{
  Application->HelpContext(3);
}
//---------------------------------------------------------------------------


//---------------------------------------------------------------------------
// Make the buttons line up when the form changes size
//
void __fastcall TformGPGeeVerifyDecrypt::FormResize(TObject *Sender)
{
  btnOkCancel->Left = Width / 2 - btnOkCancel->Width / 2;
}  // void __fastcall TformGPGeeVerifyDecrypt::FormResize(TObject *Sender)
//---------------------------------------------------------------------------


//---------------------------------------------------------------------------
// This is where the real work of getting a passphrase is done.  It's easier
// to do this from inside the class, so the callback passes control here.
//
const char * __fastcall TformGPGeeVerifyDecrypt::GetPassphrase(const char *desc)
{
  static AnsiString Passphrase = "";
  gpgme_key_t Key;
  AnsiString KeyId, KeyName, KeyType, KeySize, KeyCreationDate;

  if (desc) {
    TStringList *desclines = new TStringList;
    desclines->Text = desc;
    formGPGeePassPhrase = new TformGPGeePassPhrase(NULL);

    if (desclines->Strings[1] == "[User ID hint missing]") {
      if (formGPGeePassPhrase->GetPassphrase("Enter passphrase to decrypt symmetrically encrypted file:\r\n" + ExtraPassphraseInfo, "", "", "", "") == mrOk)
        Passphrase = formGPGeePassPhrase->edtPassphrase->Text;
      else
        Passphrase = "";
    } else if (!desclines->Strings[1].IsEmpty()) {
      KeyId = desclines->Strings[1].SubString(9,8);
      gpgme_keycache_rewind(cacheSecret);
      if (gpgme_keycache_find_key(cacheSecret,KeyId.c_str(), false, &Key) == GPGME_No_Error) {
        KeyName = String(gpgme_key_get_string_attr(Key, GPGME_ATTR_NAME, NULL, 0)) + " <" + String(gpgme_key_get_string_attr(Key, GPGME_ATTR_EMAIL, NULL, 0)) + ">";
        KeyType = String(gpgme_key_get_string_attr(Key, GPGME_ATTR_ALGO, NULL, 0));
        KeySize = String(gpgme_key_get_ulong_attr(Key, GPGME_ATTR_LEN, NULL, 0));
        KeyCreationDate = DateToStr(GPGToDateTime(gpgme_key_get_ulong_attr(Key, GPGME_ATTR_CREATED, NULL, 0)));
        if (formGPGeePassPhrase->GetPassphrase(KeyName, KeyId, KeyType, KeySize, KeyCreationDate) == mrOk)
          Passphrase = formGPGeePassPhrase->edtPassphrase->Text;
        else
          Passphrase = "";
      }  // if (gpgme_keycache_find_key(cacheSecret,KeyId.c_str(), false, &Key) == GPGME_No_Error)
    }  // if (desclines->Strings[1] == "[User ID hint missing]")
    delete formGPGeePassPhrase;
    return Passphrase.c_str();
  } else if (Passphrase.IsEmpty()) {
    // Make sure that when the passphrase is no longer needed that it is erased
    Passphrase.StringOfChar('*', Passphrase.Length());
    Passphrase = "";
  }  // if (desc)
  return NULL;
}
//---------------------------------------------------------------------------


//---------------------------------------------------------------------------
// The actual passphrase callback - just dumps control to the Passphrase
// method of the form class.
//
const char * PassphraseCallback(void *hook, const char *desc, void *dummy)
{
  TformGPGeeVerifyDecrypt *formGPGeeVerifyDecrypt = (TformGPGeeVerifyDecrypt *)hook;
  return formGPGeeVerifyDecrypt->GetPassphrase(desc);
}
//---------------------------------------------------------------------------




